@@ -55,11 +55,6 @@ struct _GsmConsolekitClass
 
         void (* request_completed) (GsmConsolekit *manager,
                                     GError        *error);
-
-        void (* privileges_completed) (GsmConsolekit *manager,
-                                       gboolean       success,
-                                       gboolean       ask_later,
-                                       GError        *error);
 };
 
 enum _GsmConsolekitError {
@@ -77,10 +72,6 @@ GsmConsolekit   *gsm_consolekit_new             (void) G_GNUC_MALLOC;
 
 gboolean         gsm_consolekit_can_switch_user (GsmConsolekit *manager);
 
-gboolean         gsm_consolekit_get_restart_privileges (GsmConsolekit *manager);
-
-gboolean         gsm_consolekit_get_stop_privileges    (GsmConsolekit *manager);
-
 gboolean         gsm_consolekit_can_stop        (GsmConsolekit *manager);
 
 gboolean         gsm_consolekit_can_restart     (GsmConsolekit *manager);
@@ -96,6 +87,8 @@ gchar           *gsm_consolekit_get_current_session_type (GsmConsolekit *manager
 
 GsmConsolekit   *gsm_get_consolekit             (void);
 
+gboolean         gsm_consolekit_is_current_session_login (void);
+
 G_END_DECLS
 
 #endif /* __GSM_CONSOLEKIT_H__ */
