@@ -57,7 +57,6 @@ struct _GsmShellPrivate
         GsmStore        *inhibitors;
 
         guint32          is_running : 1;
-        guint32          has_open_dialog : 1;
 
         DBusGProxyCall  *end_session_open_call;
         GsmShellEndSessionDialogType end_session_dialog_type;
@@ -73,8 +72,11 @@ enum {
 enum {
         END_SESSION_DIALOG_OPENED = 0,
         END_SESSION_DIALOG_OPEN_FAILED,
+        END_SESSION_DIALOG_CLOSED,
         END_SESSION_DIALOG_CANCELED,
-        END_SESSION_DIALOG_CONFIRMED,
+        END_SESSION_DIALOG_CONFIRMED_LOGOUT,
+        END_SESSION_DIALOG_CONFIRMED_SHUTDOWN,
+        END_SESSION_DIALOG_CONFIRMED_REBOOT,
         NUMBER_OF_SIGNALS
 };
 
@@ -95,6 +97,7 @@ static void     gsm_shell_on_name_owner_changed (DBusGProxy *bus_proxy,
                                                  const char *prev_owner,
                                                  const char *new_owner,
                                                  GsmShell   *shell);
+static void     queue_end_session_dialog_update (GsmShell *shell);
 
 G_DEFINE_TYPE (GsmShell, gsm_shell, G_TYPE_OBJECT);
 
@@ -159,6 +162,16 @@ gsm_shell_class_init (GsmShellClass *shell_class)
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE, 0);
 
+        signals [END_SESSION_DIALOG_CLOSED] =
+                g_signal_new ("end-session-dialog-closed",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GsmShellClass, end_session_dialog_closed),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
+
         signals [END_SESSION_DIALOG_CANCELED] =
                 g_signal_new ("end-session-dialog-canceled",
                               G_OBJECT_CLASS_TYPE (object_class),
@@ -169,11 +182,31 @@ gsm_shell_class_init (GsmShellClass *shell_class)
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE, 0);
 
-        signals [END_SESSION_DIALOG_CONFIRMED] =
-                g_signal_new ("end-session-dialog-confirmed",
+        signals [END_SESSION_DIALOG_CONFIRMED_LOGOUT] =
+                g_signal_new ("end-session-dialog-confirmed-logout",
                               G_OBJECT_CLASS_TYPE (object_class),
                               G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GsmShellClass, end_session_dialog_confirmed),
+                              G_STRUCT_OFFSET (GsmShellClass, end_session_dialog_confirmed_logout),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
+
+        signals [END_SESSION_DIALOG_CONFIRMED_SHUTDOWN] =
+                g_signal_new ("end-session-dialog-confirmed-shutdown",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GsmShellClass, end_session_dialog_confirmed_shutdown),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE, 0);
+
+        signals [END_SESSION_DIALOG_CONFIRMED_REBOOT] =
+                g_signal_new ("end-session-dialog-confirmed-reboot",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GsmShellClass, end_session_dialog_confirmed_reboot),
                               NULL,
                               NULL,
                               g_cclosure_marshal_VOID__VOID,
@@ -195,7 +228,8 @@ gsm_shell_bus_filter (DBusConnection *connection,
                                     DBUS_INTERFACE_LOCAL, "Disconnected") &&
             strcmp (dbus_message_get_path (message), DBUS_PATH_LOCAL) == 0) {
                 gsm_shell_disconnect_from_bus (shell);
-                return DBUS_HANDLER_RESULT_HANDLED;
+                /* let other filters get this disconnected signal, so that they
+                 * can handle it too */
         }
 
         return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -459,34 +493,66 @@ on_open_finished (DBusGProxy     *proxy,
                 return;
         }
 
-        shell->priv->has_open_dialog = TRUE;
         g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_OPENED], 0);
 }
 
+static void
+on_end_session_dialog_closed (DBusGProxy *proxy,
+                              GsmShell   *shell)
+{
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
+        g_signal_handlers_disconnect_by_func (shell->priv->inhibitors,
+                                              G_CALLBACK (queue_end_session_dialog_update),
+                                              shell);
+
+        g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CLOSED], 0);
+}
+
 static void
 on_end_session_dialog_canceled (DBusGProxy *proxy,
                                 GsmShell   *shell)
 {
-        if (shell->priv->update_idle_id != 0) {
-                g_source_remove (shell->priv->update_idle_id);
-                shell->priv->update_idle_id = 0;
-        }
-
-        shell->priv->has_open_dialog = FALSE;
         g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CANCELED], 0);
 }
 
 static void
-on_end_session_dialog_confirmed (DBusGProxy *proxy,
-                                 GsmShell   *shell)
+on_end_session_dialog_confirmed_logout (DBusGProxy *proxy,
+                                        GsmShell   *shell)
 {
         if (shell->priv->update_idle_id != 0) {
                 g_source_remove (shell->priv->update_idle_id);
                 shell->priv->update_idle_id = 0;
         }
 
-        shell->priv->has_open_dialog = FALSE;
-        g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CONFIRMED], 0);
+        g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CONFIRMED_LOGOUT], 0);
+}
+
+static void
+on_end_session_dialog_confirmed_shutdown (DBusGProxy *proxy,
+                                          GsmShell   *shell)
+{
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
+        g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CONFIRMED_SHUTDOWN], 0);
+}
+
+static void
+on_end_session_dialog_confirmed_reboot (DBusGProxy *proxy,
+                                        GsmShell   *shell)
+{
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
+        g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CONFIRMED_REBOOT], 0);
 }
 
 static void
@@ -508,11 +574,6 @@ on_need_end_session_dialog_update (GsmShell *shell)
 
         shell->priv->update_idle_id = 0;
 
-        if (!shell->priv->has_open_dialog)
-                return FALSE;
-
-        shell->priv->has_open_dialog = FALSE;
-
         gsm_shell_open_end_session_dialog (shell,
                                            shell->priv->end_session_dialog_type,
                                            shell->priv->inhibitors);
@@ -569,20 +630,37 @@ gsm_shell_open_end_session_dialog (GsmShell *shell,
                                   G_CALLBACK (on_end_session_dialog_proxy_destroyed),
                                   shell);
 
+                dbus_g_proxy_add_signal (shell->priv->end_session_dialog_proxy,
+                                         "Closed", G_TYPE_INVALID);
+                dbus_g_proxy_connect_signal (shell->priv->end_session_dialog_proxy,
+                                             "Closed",
+                                             G_CALLBACK (on_end_session_dialog_closed),
+                                             shell, NULL);
+
                 dbus_g_proxy_add_signal (shell->priv->end_session_dialog_proxy,
                                          "Canceled", G_TYPE_INVALID);
                 dbus_g_proxy_connect_signal (shell->priv->end_session_dialog_proxy,
                                              "Canceled",
                                              G_CALLBACK (on_end_session_dialog_canceled),
                                              shell, NULL);
-
                 dbus_g_proxy_add_signal (shell->priv->end_session_dialog_proxy,
-                                         "Confirmed", G_TYPE_INVALID);
+                                         "ConfirmedLogout", G_TYPE_INVALID);
                 dbus_g_proxy_connect_signal (shell->priv->end_session_dialog_proxy,
-                                             "Confirmed",
-                                             G_CALLBACK (on_end_session_dialog_confirmed),
+                                             "ConfirmedLogout",
+                                             G_CALLBACK (on_end_session_dialog_confirmed_logout),
+                                             shell, NULL);
+                dbus_g_proxy_add_signal (shell->priv->end_session_dialog_proxy,
+                                         "ConfirmedShutdown", G_TYPE_INVALID);
+                dbus_g_proxy_connect_signal (shell->priv->end_session_dialog_proxy,
+                                             "ConfirmedShutdown",
+                                             G_CALLBACK (on_end_session_dialog_confirmed_shutdown),
+                                             shell, NULL);
+                dbus_g_proxy_add_signal (shell->priv->end_session_dialog_proxy,
+                                         "ConfirmedReboot", G_TYPE_INVALID);
+                dbus_g_proxy_connect_signal (shell->priv->end_session_dialog_proxy,
+                                             "ConfirmedReboot",
+                                             G_CALLBACK (on_end_session_dialog_confirmed_reboot),
                                              shell, NULL);
-
         }
 
         inhibitor_array = get_array_from_store (inhibitors);
@@ -611,7 +689,17 @@ gsm_shell_open_end_session_dialog (GsmShell *shell,
                 return FALSE;
         }
 
-        shell->priv->inhibitors = g_object_ref (inhibitors);
+        g_object_ref (inhibitors);
+
+        if (shell->priv->inhibitors != NULL) {
+                g_signal_handlers_disconnect_by_func (shell->priv->inhibitors,
+                                                      G_CALLBACK (queue_end_session_dialog_update),
+                                                      shell);
+                g_object_unref (shell->priv->inhibitors);
+        }
+
+        shell->priv->inhibitors = inhibitors;
+
         g_signal_connect_swapped (inhibitors, "added",
                                   G_CALLBACK (queue_end_session_dialog_update),
                                   shell);
